/* 
 * Author: Andrei Grigorean, University of Bucharest
 * Date: 11.05.2009
 * Time complexity: O(N^3)
 * Memory complexity: O(N^2)
 * Expected score: 100/100
 */

#include <cstring>
#include <cstdlib>
#include <fstream>
#include <string>
#include <sstream>
#include <algorithm>
#include <queue>
using namespace std;

#define x first
#define y second

const int NR_TESTS = 20;
const int MAX_N = 64;

const int dx[] = {-1, 0, 1, 0};
const int dy[] = {0, 1, 0, -1};
const int dir[] = {2, 3, 0, 1};
const int dirEnd[] = {0, 1, 1, 2, 3, 0, 3, 2};
const int dirEnd2[] = {0, 1, 2, 1, 0, 3, 3, 2, 1};

int n, m;
int a[MAX_N][MAX_N];
pair<int, int> p0;
pair<int, int> pos[MAX_N * MAX_N];
bool mark[MAX_N][MAX_N];
int vis[MAX_N][MAX_N];
int t[MAX_N][MAX_N];
int stp;
ifstream *fin;
ofstream *fout;
queue< pair<int, int> > q;

void getPositions() {
	for (int i = 1; i <= n; ++i)
		for (int j = 1; j <= m; ++j)
			pos[ a[i][j] ] = make_pair(i, j);
}

inline bool validPos(pair<int, int> p) {
	if (p.x < 1 || p.x > n) return false;
	if (p.y < 1 || p.y > m) return false;
	return !mark[p.x][p.y];
}

inline int dist(pair<int, int> p1, pair<int, int> p2) {
	return abs(p1.x - p2.x) + abs(p1.y - p2.y);
}

void printMatrix() {
	(*fout) << "\n" << "\n";
	for (int i = 1; i <= n; ++i) {
		for (int j = 1; j <= m; ++j)
			(*fout) << a[i][j] << " ";
		(*fout) << "\n";
	}
	(*fout) << "\n";
}


void makeMove(int k) {
	(*fout) << k + 1 << " ";

	int x = pos[0].x + dx[k];
	int y = pos[0].y + dy[k];
	int val = a[x][y];

	swap(a[x][y], a[pos[0].x][pos[0].y]);
	swap(pos[val], pos[0]);
}

void move0(pair<int, int> dest) {
	while (!q.empty()) q.pop();
	q.push(dest);

	vis[dest.x][dest.y] = ++stp;

	while (!q.empty() && vis[pos[0].x][pos[0].y] != stp) {
		pair<int, int> p = q.front();
		q.pop();

		for (int k = 0; k < 4; ++k) {
			pair<int, int> p2 = make_pair(p.x + dx[k], p.y + dy[k]);
			if (!validPos(p2) || vis[p2.x][p2.y] == stp) continue;

			t[p2.x][p2.y] = dir[k];
			vis[p2.x][p2.y] = stp;
			q.push(p2);
		}
	}

	while (pos[0] != dest)
		makeMove(t[pos[0].x][pos[0].y]);
}

void move(int val, pair<int, int> dest) {
	while (pos[val] != dest) {
		for (int k = 0; k < 4; ++k) {
			pair<int, int> p = make_pair(pos[val].x + dx[k], pos[val].y + dy[k]);
			if (!validPos(p) || dist(p, dest) >= dist(pos[val], dest)) continue;

			mark[pos[val].x][pos[val].y] = true;
			move0(p);
			mark[pos[val].x][pos[val].y] = false;

			makeMove(dir[k]);
		}
	}
}

bool isOk() {
	if (pos[0] != make_pair(n, m)) return false;
	if (pos[n * m - 1] != make_pair(n, m - 1)) return false;
	if (pos[(n - 1) * m] != make_pair(n - 1, m)) return false;
	if (pos[(n - 1) * m - 1] != make_pair(n - 1, m - 1)) return false;
	return true;
}

int main() {
	for (int test = 1; test <= NR_TESTS; ++test) {
		ostringstream filein;
		filein << test << "-perspico.in";
		fin = new ifstream(filein.str().c_str());

		ostringstream fileout;
		fileout << test << "-perspico.ok";
		fout = new ofstream(fileout.str().c_str());

		memset(mark, false, sizeof(mark));

		(*fin) >> n >> m;
		for (int i = 1; i <= n; ++i)
			for (int j = 1; j <= m; ++j)
				(*fin) >> a[i][j];

		getPositions();
		p0 = pos[0];

		for (int i = 1; i + 2 <= n; ++i) {
			for (int j = 1; j + 1 <= m; ++j) {
				move(j + (i - 1) * m, make_pair(i, j));
				mark[i][j] = true;
			}

			if (pos[i * m] == make_pair(i, m)) {
				mark[i][m] = true;
				continue;
			}

			move(i * m, make_pair(i + 1, m));
			if (pos[0] == make_pair(i, m)) makeMove(2);
			else {
				mark[i + 1][m] = true;
				move0(make_pair(i + 1, m - 2));
				mark[i + 1][m] = false;
				for (int k = 0; k < sizeof(dirEnd)/sizeof(dirEnd[0]); ++k)
					makeMove(dirEnd[k]);
			}
			mark[i][m] = true;
		}

		for (int j = 1; j + 2 <= m; ++j) {
			int v1 = (n - 2) * m + j, v2 = (n - 1) * m  + j;

			move(v1, make_pair(n - 1, j));
			mark[n - 1][j] = true;

			if (pos[v2] == make_pair(n, j)) {
				mark[n][j] = true;
				continue;
			}

			if (pos[v2] == make_pair(n, j + 1) && pos[0] == make_pair(n, j)) {
				makeMove(1);
				mark[n][j] = true;
				continue;
			}

			move(v2, make_pair(n, j + 2));
			mark[n][j + 2] = true;

			move0(make_pair(n, j));

			mark[n - 1][j] = mark[n][j + 2] = false;
			for (int k = 0; k < sizeof(dirEnd2)/sizeof(dirEnd2[0]); ++k)
				makeMove(dirEnd2[k]);
			mark[n - 1][j] = mark[n][j] = true;
		}

		//int cnt = 0;
		while (!isOk()) {
			if (pos[0] == make_pair(n, m - 1)) makeMove(1);
			else if (pos[0] == make_pair(n, m)) makeMove(0);
			else if (pos[0] == make_pair(n - 1, m)) makeMove(3);
			else if (pos[0] == make_pair(n - 1, m - 1)) makeMove(2);
			//if (++cnt == 100) break;
		}
		//if (cnt == 100) (*fout) << "\n\n-1\n";
		//else printMatrix();
	}
}